/*
 * Back-end main module
 * See "gdiff.h" for the details of data structure.
 * This module should be independent from GUI frontend.
 *
 * Programmed by INOUE Seiichiro <inoue@ainet.or.jp>
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#if defined(HAVE_ERRNO_H)
#include <errno.h>
#else defined(HAVE_SYS_ERRNO_H)
#include <sys/errno.h>
#endif
#include <glib.h>
#include "gdiff.h"


/* Private function declarations */
static DiffFiles* new_diff_files(const char *fname1, const char *fname2);
static void delete_diff_files(DiffFiles *dfiles);
static void init_fileinfo(FileInfo *fi, char const *fname);
static void term_fileinfo(FileInfo *fi);
static int get_nlines(const char *text, int lenb);

/**
 * new_diff_dir:
 * Allocate DiffDir structure, and return its pointer.
 * Input:
 * const char *fname1; Regular file or directory.
 * const char *fname2; Regular file or directory.
 * Output:
 * Return value: DiffDir*;
 */
DiffDir*
new_diff_dir(const char *fname1, const char *fname2)
{
	DiffDir *diffdir;

	diffdir = g_new(DiffDir, 1);
	diffdir->dfiles_list = NULL;
	add_files_to_diff_dir(diffdir, fname1, fname2);
	run_diff(diffdir);
	
	return diffdir;
}

/**
 * delete_diff_dir:
 * Finalize DiffDir structure, and free its memory.
 * Input:
 * DiffDir *diffdir;
 * Output:
 * None;
 */
void
delete_diff_dir(DiffDir *diffdir)
{
	GSList *list;
	
	for (list = diffdir->dfiles_list; list; list = list->next) {
		delete_diff_files((DiffFiles*)(list->data));
	}
	g_slist_free(diffdir->dfiles_list);
	g_free(diffdir);
}

/**
 * add_files_to_diff_dir:
 * Allocate DiffFiles(call new_diff_files()), 
 * and add it to GSList *dfiles_list of DiffDir structure.
 * Never move the first node, which is a special node.
 * On the other hand, the other nodes' order doesn't matter.
 * Input:
 * DiffDir *diffdir;
 * const char *fname1;
 * const char *fname2;
 * Output:
 * DiffDir *diffdir; GSList *dfiles_list is updated.
 * Return value: DiffFiles*; Added DiffFiles' pointer.
 */
DiffFiles*
add_files_to_diff_dir(DiffDir *diffdir, const char *fname1, const char *fname2)
{
	DiffFiles *dfiles;

	dfiles = new_diff_files(fname1, fname2);
	/* append is not so efficient, but don't care */
	diffdir->dfiles_list = g_slist_append(diffdir->dfiles_list, dfiles);

	return dfiles;
}

/**
 * get_fileinfo_from_diff_files:
 * Return the pointer to FileInfo(member variable of DiffFiles).
 * If the caller needs an actual text content or its number of lines,
 * it should set TRUE to f_mmap.
 * mmap() has been delayed, so if "f_mmap" is TRUE, do mmap() here.
 * Input:
 * DiffFiles *dfiles;
 * WhichFile n;
 * gboolean f_mmap; if TRUE do mmap its text, unless it has done.
 * Output:
 * DiffFiles *dfiles; "text" and "nlines" can be updated by mmap().
 * Return value; pointer to FileInfo.
 */
const FileInfo*
get_fileinfo_from_diff_files(DiffFiles *dfiles, WhichFile n, gboolean f_mmap)
{
	FileInfo *fi = &dfiles->fileinfo[n];
	int fd;

	if (f_mmap == TRUE
		&& fi->f_dir == FALSE && fi->text == NULL && fi->lenb > 0) {
		if ((fd = open(fi->fname, O_RDONLY)) == -1) {
			g_warning("get_fileinfo_from_diff_files:open %s %d", fi->fname, errno);
			return fi;
		}
		/*XXX: MAP_NOEXTEND is dependent on platform */
		/*fi->text = mmap(0, fi->lenb, PROT_READ, MAP_PRIVATE|MAP_NOEXTEND, fd, 0);*/
		fi->text = mmap(0, fi->lenb, PROT_READ, MAP_PRIVATE, fd, 0);
		if (fi->text == (caddr_t)-1) {
			g_warning("get_fileinfo_from_diff_files:mmap %s %d", fi->fname, errno);
			fi->text = NULL;
			return fi;
		}
		fi->nline = get_nlines(fi->text, fi->lenb);
		close(fd);
	}
	return fi;
}

/**
 * add_dlines_to_diff_files:
 * Allocate DiffLines, and add it to GSList *dlines_list of DiffFiles structure. 
 * Input:
 * DiffFiles *dfiles;
 * DiffType dtype;
 * int f1n1, f1n2, f2n1, f2n2; Each is line number.
 * Output:
 * DiffFiles* dfiles; GSList *dlines_list is updated.
 */
void
add_dlines_to_diff_files(DiffFiles* dfiles, DiffType dtype, int f1n1, int f1n2, int f2n1, int f2n2)
{
	DiffLines *dlines;
	
	dlines = g_new(DiffLines, 1);
	dlines->difftype = dtype;
	dlines->between[FIRST_FILE].begin = f1n1;
	dlines->between[FIRST_FILE].end = f1n2;
	dlines->between[SECOND_FILE].begin = f2n1;
	dlines->between[SECOND_FILE].end = f2n2;

	/* Not so efficient, but don't care */
	dfiles->dlines_list = g_slist_append(dfiles->dlines_list, dlines);
}



/* ---The followings are private functions--- */
/**
 * new_diff_files:
 * Allocate DiffFiles structure, initialize and return its pointer.
 * Input:
 * const char *fname1; First file name. Directory is acceptable.
 * const char *fname2; Second file name. Directory is acceptable.
 * Output:
 * Return value: DiffFile*; pointer to the allocated DiffFiles structure;
 */
static DiffFiles*
new_diff_files(const char *fname1, const char *fname2)
{
	DiffFiles *dfiles;

	dfiles = g_new(DiffFiles, 1);

	/* In the case only one file exists, don't open the file. */
	if (fname1[0] == '\0' || fname2[0] == '\0') {
		init_fileinfo(&dfiles->fileinfo[FIRST_FILE], NULL);
		init_fileinfo(&dfiles->fileinfo[SECOND_FILE], NULL);
		if (fname1[0])
			dfiles->fileinfo[FIRST_FILE].fname = g_strdup(fname1);
		if (fname2[0])
			dfiles->fileinfo[SECOND_FILE].fname = g_strdup(fname2);
	} else {
		init_fileinfo(&dfiles->fileinfo[FIRST_FILE], fname1);		
		init_fileinfo(&dfiles->fileinfo[SECOND_FILE], fname2);
	}
	dfiles->dlines_list = NULL;
	dfiles->binary = FALSE;

	return dfiles;
}


/**
 * delete_diff_files:
 * Finalize DiffFiles structure, and free its memory.
 * Input:
 * DiffFiles *dfiles;
 * Output:
 * None;
 */
static void
delete_diff_files(DiffFiles *dfiles)
{
	GSList *list;
	
	term_fileinfo(&dfiles->fileinfo[FIRST_FILE]);
	term_fileinfo(&dfiles->fileinfo[SECOND_FILE]);

	for (list = dfiles->dlines_list; list; list = list->next) {
		g_free(list->data);
	}
	g_slist_free(dfiles->dlines_list);

	g_free(dfiles);
}


/**
 * init_fileinfo:
 * Initialize FileInfo structure.
 * But it doesn't do mmap(), and make it delayed for get_fileinfo_from_diff_files().
 * Input:
 * const char *fname;
 * Output:
 * FileInfo *fi;
 */
static void
init_fileinfo(FileInfo *fi, const char *fname)
{
	int fd;
	struct stat sb;

	/* default values */
	fi->nline = 0;
	fi->text = NULL;
	fi->lenb = 0;
	fi->f_dir = FALSE;

	if (fname) {
		fi->fname = g_strdup(fname);
		if ((fd = open(fname, O_RDONLY)) == -1) {
			g_warning("init_fileinfo:open %s %d", fname, errno);
			return;
		}
		if (fstat(fd, &sb) == -1) {
			g_warning("init_fileinfo:fstat %s %d", fname, errno);
			close(fd);
			return;
		}
		if (sb.st_mode & S_IFREG) {
			fi->lenb = sb.st_size;
		}
		else if (sb.st_mode & S_IFDIR) {
			fi->f_dir = TRUE;
		}
		else {
			/* not regular-file, nor directory. */
			g_warning("init_fileinfo:special file");
		}
		close(fd);
	} else {
		fi->fname = NULL;
	}
}

/**
 * term_fileinfo:
 * Terminate FileInfo structure. (not free its own memory.)
 * Input:
 * FileInfo *fi;
 * Output:
 * None;
 */
static void
term_fileinfo(FileInfo *fi)
{
	if (fi->fname) {
		g_free(fi->fname);
	}
	if (fi->text) {
		if (munmap(fi->text, fi->lenb) == -1)
			g_warning("munmap in term_fileinfo %d", errno);
	}
}


/**
 * get_nlines:
 * Return the number of lines of the buffer.
 * Counts the last line, although it isn't ended with '\n' character.
 * Input:
 * const char *text; Buffer.
 * int lenb; Buffer length(Bytes).
 * Output:
 * Return value; The number of lines of the buffer.
 */
static int
get_nlines(const char *text, int lenb)
{
	int nl = 0;
	const char *p = text;
	int i;
	
	for (i = 0; i < lenb; i++) {
		if (*p == '\n')
			nl++;
		p++;
	}
	if (p != text && *(p-1) != '\n')
		nl++;

	return nl;
}
