/*
 * Back-end misc module
 * See "gdiff.h" for the details of data structure.
 * This module should be independent from GUI frontend.
 *
 * Programmed by INOUE Seiichiro <inoue@ainet.or.jp>
 */
#include <glib.h>
#include "gdiff.h"


/**
 * get_files_special_status:
 * See "gdiff.h" about enum FilesSpecialStatus.
 * Input:
 * const DiffFiles *dfiles;
 * Output:
 * Return value: FilesSpecialStatus.
 */
FilesSpecialStatus
get_files_special_status(DiffFiles *dfiles)
{
	const FileInfo *fi1 = get_fileinfo_from_diff_files(dfiles, FIRST_FILE, FALSE);
	const FileInfo *fi2 = get_fileinfo_from_diff_files(dfiles, SECOND_FILE, FALSE);
	
	if (dfiles->binary)
		return BINARY_FILES;
	if (fi1->fname == NULL) 
		return ONLY_FILE2_EXISTS;
	if (fi2->fname == NULL)
		return ONLY_FILE1_EXISTS;
	if (fi1->f_dir == TRUE) {
		/* double check */
		if (fi2->f_dir == FALSE) {
			g_warning("something wrong in get_files_special_status()\n");
		}
		return DIRECTORIES;
	}
	if (dfiles->dlines_list == NULL)
		return IDENTICAL_FILES;

	return DIFFERENT_FILES;
}

/**
 * get_merged_numlines:
 * Return the number of lines of merged buffers.
 * Input:
 * DiffFiles *files;
 * Output:
 * DiffFiles *files; Could be updated (by calling get_fileinfo_from_diff_files()). But theoretically, not updated here.
 * Return value; the number of lines of merged buffers.
 */
int
get_merged_numlines(DiffFiles *files)
{
	const FileInfo *fi1 = get_fileinfo_from_diff_files(files, FIRST_FILE, TRUE);
	int merged_nline;
	GSList *node;/* node of DiffLines' list */
	DiffLines *dlines;
	
	merged_nline = fi1->nline;/* first file */
	/* second file */
	node = files->dlines_list;
	while (node) {
		dlines = (DiffLines*)(node->data);
		if (dlines->difftype == F2ONLY || dlines->difftype == CHANGE) {
			merged_nline += (dlines->between[SECOND_FILE].end - dlines->between[SECOND_FILE].begin + 1);
		}
		node = g_slist_next(node);
	}
	
	return merged_nline;
}

/**
 * calc_diff_total_lines:
 * Calculate the tolal number of different lines, and return it.
 * Input:
 * const DiffFiles *dfiles;
 * WhichFile n;
 * Output:
 * Return value; The tolal number of different lines.
 */
int
calc_diff_total_lines(const DiffFiles *dfiles, WhichFile n)
{
	GSList *list;/* Singly linked list of DiffLines */
	int ret_nline = 0;
	
	for (list = dfiles->dlines_list; list; list = list->next) {
		DiffLines *dlines = (DiffLines*)list->data;
		ret_nline += dlines->between[n].end - dlines->between[n].begin + 1; 
	}
	return ret_nline;
}


/**
 * get_max_nlines:
 * Compare the number of lines of two files, and return one of the bigger.
 * Input:
 * const DiffFiles *dfiles;
 * Output:
 * Return value; The number of lines of the bigger.
 */
int
get_max_nlines(DiffFiles *dfiles)
{
	const FileInfo *fi1 = get_fileinfo_from_diff_files(dfiles, FIRST_FILE, TRUE);
	const FileInfo *fi2 = get_fileinfo_from_diff_files(dfiles, SECOND_FILE, TRUE);

	return (fi1->nline > fi2->nline) ? fi1->nline : fi2->nline;
}
