/*
 * Module for directory view
 *
 * Programmed by INOUE Seiichiro <inoue@ainet.or.jp>
 */
#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include "gdiff.h"
#include "gui.h"

/* Constant numbers */
enum {
	COL_FILENAME1 = 0,
	COL_FILENAME2 = 1,
	COL_TYPE = 2,
	COL_NLINES1 = 3,
	COL_NLINES2 = 4,
};

/* Private data structure definitions */
typedef struct {
	char *title;
	GtkCListCompareFunc cmp_func;
} DirViewCList;

/* Private function declarations */
static void
create_clist_dir(GDiffDirViews *gdirviews, GtkWidget *hbox);
static void click_dirview(GtkCList *clist, gint row, gint column, GdkEvent *event, gpointer data);
static void click_column(GtkCList *clist, gint column, gpointer data);
static gint number_compare(GtkCList *clist, gconstpointer ptr1, gconstpointer ptr2);

/* Private variables */
static DirViewCList dview_clist[] = {
	{"filename1", NULL},
	{"filename2", NULL},
	{"diff-type", NULL},
	{"diff lines1", number_compare},
	{"diff lines2", number_compare},
};
/*static const int DVIEW_COLUMNS = sizeof(dview_clist) / sizeof(dview_clist[0]);*/
#define DVIEW_COLUMNS	5


/**
 * create_dir_widgets:
 * Called from gdiff_dirviews_new().
 * Create widgets and show them.
 * The contents(diff results) are not shown here, which up to display_dirs().
 * Input:
 * GDiffDirViews *gdirviews;
 * Output:
 * GDiffDirViews *gdirviews; updated.
 */
void
create_dir_widgets(GDiffDirViews *gdirviews)
{
	GDiffWindow *gdwin = gdirviews->gdwin;
	GtkWidget *hbox;
	GtkWidget *label;
	gint page;

	/* hbox */
	hbox = gtk_hbox_new(FALSE, 0);
	gdirviews->hbox = GTK_HBOX(hbox);
	gdirviews->base = GTK_WIDGET(gdirviews->hbox);/* Virtual base widget. */
	
	/* Add to the notebook */
	label = make_label_for_notebook(gdirviews->dirname1, gdirviews->dirname2);
	gdirviews->label = GTK_LABEL(label);
	gtk_notebook_append_page(gdwin->notebook, hbox, label);

	/* clist area */
	create_clist_dir(gdirviews, hbox);

	gtk_widget_show(hbox);
	/* Is this a proper way ? */
	page = gtk_notebook_page_num(gdwin->notebook, hbox);
	gtk_notebook_set_page(gdwin->notebook, page);
}


/**
 * display_dirs:
 * Display diff directory, i.e. file names list, on GtkCList,
 * getting the information from back-end data(DiffDir *diffdir).
 * If comparing two files, call a function to display diff files.
 * (See "gdiff.h", I deal with comparing two files as one case.)
 * Input:
 * GDiffDirViews *gdirviews;
 * Outupt:
 * None;
 */
void
display_dirs(GDiffDirViews *gdirviews)
{
#define CAPTION_DIFFERENT	"Different"
#define CAPTION_BINARY		"Binary"
#define CAPTION_ONLYONE		"Only"
	if (gdirviews->isdir) {
		DiffDir *dir = gdirviews->diffdir;
		GSList *list;
		GtkCList *clist = gdirviews->clist;
		int row = 0;
		
		gtk_clist_freeze(clist);

		/* ignore the first node, which has directory names. */
		for (list = dir->dfiles_list->next; list; list = list->next) {
			DiffFiles *files = (DiffFiles*)list->data;
			char *rowtext[DVIEW_COLUMNS];
			FilesSpecialStatus fsst;
			int n;
			char buf1[32];
			char buf2[32];
			const FileInfo *fi1 = get_fileinfo_from_diff_files(files, FIRST_FILE, FALSE);
			const FileInfo *fi2 = get_fileinfo_from_diff_files(files, SECOND_FILE, FALSE);

			rowtext[COL_FILENAME1] = fi1->fname;
			rowtext[COL_FILENAME2] = fi2->fname;
			fsst = get_files_special_status(files);
			switch (fsst) {
			case IDENTICAL_FILES:
			case DIRECTORIES:
				g_warning("XXX: weird in display_dirs %d\n", fsst);
				return;
			case BINARY_FILES:
				rowtext[COL_TYPE] = CAPTION_BINARY;
				break;
			case ONLY_FILE1_EXISTS:
			case ONLY_FILE2_EXISTS:
				rowtext[COL_TYPE] = CAPTION_ONLYONE;
				break;
			case DIFFERENT_FILES:
				rowtext[COL_TYPE] = CAPTION_DIFFERENT;
				break;
			}
			n = calc_diff_total_lines(files, FIRST_FILE);
			sprintf(buf1, "%d", n);
			rowtext[COL_NLINES1] = buf1;
			n = calc_diff_total_lines(files, SECOND_FILE);
			sprintf(buf2, "%d", n);
			rowtext[COL_NLINES2] = buf2;
			gtk_clist_append(clist, rowtext);
			/* Kludge: take care of the case you change the row order. */
			gtk_clist_set_row_data(GTK_CLIST(clist), row, (gpointer)row);
			row++;
		}
		gtk_signal_connect(GTK_OBJECT(clist), "select_row",
						   GTK_SIGNAL_FUNC(click_dirview), gdirviews);
		gtk_signal_connect(GTK_OBJECT(clist), "click_column",
						   GTK_SIGNAL_FUNC(click_column), gdirviews);
		gtk_clist_thaw(clist);
	} else {
		GDiffFileViews *gfviews;

		gfviews = g_slist_nth_data(gdirviews->gdfile_list, 0);
		if (gfviews->state.pane_mode == ONE_PANE)
			one_pane_display(gfviews);
		else if (gfviews->state.pane_mode == TWO_PANE)
			two_pane_display(gfviews);
	}
}



/* ---The followings are private functions--- */
/**
 * create_clist_dir:
 * Create GtkCList widget for directory list.
 * Input:
 * GDiffDirViews *gdirviews;
 * GtkWidget *hbox; Parent of GtkCList.
 * Output:
 * None;
 */
static void
create_clist_dir(GDiffDirViews *gdirviews, GtkWidget *hbox)
{
	GtkWidget *clist;
	GtkWidget *scrolled_window;
	int n;
	char *titles[DVIEW_COLUMNS];


	for (n = 0; n < DVIEW_COLUMNS; n++) {
		titles[n] = dview_clist[n].title;
	}
	scrolled_window = gtk_scrolled_window_new(NULL, NULL);
	gtk_box_pack_start(GTK_BOX(hbox),
					   scrolled_window, TRUE, TRUE, 0);
	gtk_widget_show(scrolled_window);

	clist = gtk_clist_new_with_titles(DVIEW_COLUMNS, titles);
	gtk_clist_column_titles_active(GTK_CLIST(clist));
	/* TODO: column width */
	gtk_clist_set_column_width(GTK_CLIST(clist), COL_FILENAME1, 150);
	gtk_clist_set_column_width(GTK_CLIST(clist), COL_FILENAME2, 150);
	gtk_container_add(GTK_CONTAINER(scrolled_window), clist);
	gtk_widget_grab_focus(clist);
	gtk_widget_show(clist);

	gdirviews->clist = GTK_CLIST(clist);
}

/**
 * click_dirview:
 * Called when a user click the row on directory view.
 * Open the files view(diff results) for them.
 * At first, look for the opened view.
 * If the same mode view as current setting exists, just focus it.
 * If the same mode view doesn't exist, open a new view in the current mode.
 */
static void
click_dirview(GtkCList *clist, gint row, gint column, GdkEvent *event, gpointer data)
{
	GDiffDirViews *gdviews = (GDiffDirViews*)data;
	GDiffFileViews *gfviews;
	DiffDir *diffdir = gdviews->diffdir;
	DiffFiles *dfiles;
	GSList *list;
	FilesSpecialStatus fsst;
	int num;

	
	/* Kludge: take care of the case you change the row order. */
	num = (int)gtk_clist_get_row_data(clist, row);
	dfiles = g_slist_nth_data(diffdir->dfiles_list, num+1);
	fsst = get_files_special_status(dfiles);
	switch (fsst) {
	case IDENTICAL_FILES:
		g_message("identical");
		return;
	case BINARY_FILES:
		g_message("binary");
		return;
	case ONLY_FILE1_EXISTS:
	case ONLY_FILE2_EXISTS:
		g_message("The file exists only in one directory\n");
		return;
	default:
		/* through */
		break;
	}

	/* Look for opened views */
	for (list = gdviews->gdfile_list; list; list = list->next) {
		gfviews = (GDiffFileViews*)list->data;
		if (gfviews->dfiles == dfiles) {/* Opened */
			gint page;
			page = gtk_notebook_page_num(gfviews->gdwin->notebook, gfviews->base);
			/* this is the same mode as current setting. */
			if (gfviews->state.pane_mode == g_pref.pane_mode) {
				if (!GTK_WIDGET_VISIBLE(gfviews->base)) {
					gtk_widget_show(gfviews->base);
				}
				gtk_notebook_set_page(gfviews->gdwin->notebook, page);
				return;
			}
		}
	}

	/* When not opened, make it open */	
	gfviews = gdiff_fileviews_new(gdviews->gdwin, gdviews, dfiles, g_pref.pane_mode);
	if (gfviews->state.pane_mode == ONE_PANE)
		one_pane_display(gfviews);
	else
		two_pane_display(gfviews);
}


/**
 * click_column:
 */
static void
click_column(GtkCList *clist, gint column, gpointer data)
{
	if (column == clist->sort_column) {
		if (clist->sort_type == GTK_SORT_ASCENDING)
			clist->sort_type = GTK_SORT_DESCENDING;
		else
			clist->sort_type = GTK_SORT_ASCENDING;
    } else {
		gtk_clist_set_sort_column(clist, column);
		gtk_clist_set_compare_func(clist, dview_clist[column].cmp_func);
	}
	gtk_clist_sort (clist);
}     

/**
 * number_compare:
 * Compare function used for sort of numbers.
 * I modified default_compare() in gtkclist.c(GTK+1.2.0).
 */
static gint
number_compare(GtkCList *clist, gconstpointer ptr1, gconstpointer ptr2)
{
	int num1;
	int num2;
	char *text1 = NULL;
	char *text2 = NULL;
	GtkCListRow *row1 = (GtkCListRow *) ptr1;
	GtkCListRow *row2 = (GtkCListRow *) ptr2;

	switch (row1->cell[clist->sort_column].type) {
    case GTK_CELL_TEXT:       
		text1 = GTK_CELL_TEXT(row1->cell[clist->sort_column])->text;
		break;
    case GTK_CELL_PIXTEXT:    
		text1 = GTK_CELL_PIXTEXT(row1->cell[clist->sort_column])->text; 
		break;
    default:
		break;
    }

	switch (row2->cell[clist->sort_column].type) {
    case GTK_CELL_TEXT:       
		text2 = GTK_CELL_TEXT(row2->cell[clist->sort_column])->text;
		break;
    case GTK_CELL_PIXTEXT:    
		text2 = GTK_CELL_PIXTEXT(row2->cell[clist->sort_column])->text; 
		break;
    default:
		break;
    }

	if (!text2)
		return (text1 != NULL);   
	if (!text1)
		return -1;

	num1 = atoi(text1);
	num2 = atoi(text2);

	return num1 - num2;
}
