/*
 * GUI file operation module
 * Mainly, handling GtkFileSelection widget.
 *
 * Programmed by INOUE Seiichiro <inoue@ainet.or.jp>
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <gtk/gtk.h>
#include "gdiff.h"
#include "gui.h"
#include "misc.h"


/* Note:
 * File dialog box has a state, because it has to accept two file names.*/

/* Constant strings */
#define FIRST_FILE_SEL_CAPTION		"First file selection"
#define SECOND_FILE_SEL_CAPTION		"Second file selection"


/* Data structure definitions */

/* Data kept by GtkFileSelection instance.
 * accessed through  gtk_object_set_user_data() and gtk_object_get_user_data() interfaces. */
typedef struct {
	/* For reference */
	GDiffWindow *gdwin;

	/* Dialog box state */
	gboolean is_directory;
	char *fname1;
	GtkLabel *fname1_label;
	WhichFile file_count;
} FileDialogKeep;


/* Private function declarations */
static void files_selection_dialog(GDiffWindow *gdwin);
static void file_selection_ok(GtkWidget *w, gpointer data);
static void file_selection_cancel(GtkWidget *w, gpointer data);
static void dir_toggle(GtkWidget *w, gpointer data);
static FileDialogKeep* filedialogkeep_new(GDiffWindow *gdwin, GtkLabel *label);
static void filedialogkeep_reset(FileDialogKeep *fdk);
static void filedialogkeep_update(FileDialogKeep *fdk, const char *fname1);


/**
 * files_open:
 * An entry point to open files.
 * Now, called only from [menu]-[File]-[Open].
 */
void
files_open(GDiffWindow *gdwin)
{
	files_selection_dialog(gdwin);
}


/* ---The followings are private functions--- */
/**
 * files_selection_dialog:
 * If the instance of dialog has not created yet, create it and show it.
 * If it exists and has been hidden, just show it.
 * If it has been shown, do nothing.(should focus it?)
 * Input:
 * GDiffWindow *gdwin;
 * Output:
 * None;
 */
static void
files_selection_dialog(GDiffWindow *gdwin)
{
	FileDialogKeep *fdk;
	GtkWidget *filesel;
	GtkWidget *checkbutton;
	GtkWidget *label;

	/* The dialog has already been created */
	if (gdwin->filesel) {
		if (!GTK_WIDGET_VISIBLE(gdwin->filesel)) {
			fdk = gtk_object_get_user_data(GTK_OBJECT(gdwin->filesel));
			
			if (fdk->file_count == FIRST_FILE)
				gtk_window_set_title(GTK_WINDOW(gdwin->filesel),
									 FIRST_FILE_SEL_CAPTION);
			/*
			else if (fdk->file_count == SECOND_FILE)
				gtk_window_set_title(GTK_WINDOW(fs), SECOND_FILE_SEL_CAPTION);
				*/
			gtk_widget_show(GTK_WIDGET(gdwin->filesel));
		}
		/*XXX:? gtk_window_activate_focus(GTK_WINDOW(gdwin->filesel));*/
		return;
	}

	/* Create GtkFileSelection dialog */
	filesel = gtk_file_selection_new(FIRST_FILE_SEL_CAPTION);
	gtk_file_selection_hide_fileop_buttons(GTK_FILE_SELECTION(filesel));
	gtk_window_set_position(GTK_WINDOW(filesel), GTK_WIN_POS_CENTER);
	gdwin->filesel = GTK_FILE_SELECTION(filesel);

	/* Label to show first file name while you're specifing second file. */
	label = gtk_label_new("");
	gtk_box_pack_start(GTK_BOX(GTK_FILE_SELECTION(filesel)->action_area), 
					   label, FALSE, FALSE, 0);

	/* Set FileDialogKeep as user_data of file dialog widget */
	fdk = filedialogkeep_new(gdwin, GTK_LABEL(label));
	gtk_object_set_user_data(GTK_OBJECT(filesel), fdk);
	
	/*XXX: this causes an error.
	gtk_signal_connect(GTK_OBJECT(filesel), "destroy",
					   GTK_SIGNAL_FUNC(filesel_destroyed),
					   filesel);*/
	gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->ok_button),
					   "clicked", GTK_SIGNAL_FUNC(file_selection_ok), filesel);
	gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->cancel_button),
					   "clicked", GTK_SIGNAL_FUNC(file_selection_cancel), filesel);

	/* Checkbutton to specify directories */
	checkbutton = gtk_check_button_new_with_label("Compare directories");
	/* XXX:what effect? gtk_toggle_button_set_mode(GTK_TOGGLE_BUTTON(checkbutton), TRUE);*/
	gtk_signal_connect(GTK_OBJECT(checkbutton),
					   "toggled", GTK_SIGNAL_FUNC(dir_toggle), filesel);
	gtk_box_pack_start(GTK_BOX(GTK_FILE_SELECTION(filesel)->action_area), 
					   checkbutton, FALSE, FALSE, 0);
	gtk_widget_show(checkbutton);
	
    gtk_widget_show(filesel);
}

/**
 * file_selection_ok:
 * Depend on the state.
 * In the case selecting first file, change the state for second file.
 * In the case selecting second file, do the job.
 */
static void
file_selection_ok(GtkWidget *w, gpointer data)
{
	GtkFileSelection *fs = (GtkFileSelection*)data;
	FileDialogKeep *fdk;
	char *fname;
	FileType ft;

	fdk = gtk_object_get_user_data(GTK_OBJECT(fs));
	fname = gtk_file_selection_get_filename(fs);
	ft = check_filetype(fname);
	if (ft == OTHERFILE) {
		g_message("wrong file type\n");
		return;
	} else if (ft == REGULARFILE && fdk->is_directory) {
		g_message("You have to select a directory\n");
		return;
	} else if (ft == DIRECTORY && !(fdk->is_directory)) {
		g_message("You have to select a file\n");
		return;
	}
	
	if (fdk->file_count == FIRST_FILE) {
		filedialogkeep_update(fdk, fname);
		gtk_window_set_title(GTK_WINDOW(fs), SECOND_FILE_SEL_CAPTION);
		gtk_file_selection_set_filename(GTK_FILE_SELECTION(fs), "");
	} else if (fdk->file_count == SECOND_FILE) {
		GDiffWindow *gdwin;
		GDiffDirViews *gdirviews;
		char *fname1;
		char *fname2;
		
		gdwin = fdk->gdwin;
		fname1 = fdk->fname1;
		fname2 = fname;

		gdirviews = gdiff_dirviews_new(gdwin, fname1, fname2, fdk->is_directory);
		
		filedialogkeep_reset(fdk);
		gtk_widget_hide(GTK_WIDGET(fs));

		display_dirs(gdirviews);
	} else {
		g_print("all set");
	}
}

/**
 * file_selection_cancel:
 * Reset all state.
 */
static void
file_selection_cancel(GtkWidget *w, gpointer data)
{
	GtkFileSelection *fs = (GtkFileSelection*)data;
	FileDialogKeep *fdk;

	fdk = gtk_object_get_user_data(GTK_OBJECT(fs));
	filedialogkeep_reset(fdk);
	gtk_widget_hide(GTK_WIDGET(fs));
}

/**
 * dir_toggle:
 * Toggle whether selecting directories or files.
 */
static void
dir_toggle(GtkWidget *w, gpointer data)
{
	GtkFileSelection *fs = (GtkFileSelection*)data;
	FileDialogKeep *fdk;

	fdk = gtk_object_get_user_data(GTK_OBJECT(fs));
		
	if (fdk->file_count == FIRST_FILE)
		fdk->is_directory = !(fdk->is_directory);
	else /* not change */
		GTK_TOGGLE_BUTTON(w)->active = !GTK_TOGGLE_BUTTON(w)->active;/* Is this a proper way ?*/
}

/**
 * filedialogkeep_new:
 * Allocate FileDialogKeep, intialize it and return its pointer.
 * Input:
 * GDiffWindow *gdwin; kept by FileDialogKeep.
 * GtkLabel *label; kept by FileDialogKeep.
 * Output:
 * Return value; allocated FileDialogKeep pointer.
 */
static FileDialogKeep*
filedialogkeep_new(GDiffWindow *gdwin, GtkLabel *label)
{
	FileDialogKeep *fdk;

	fdk = g_new(FileDialogKeep, 1);

	fdk->gdwin = gdwin;
	fdk->is_directory = FALSE;
	fdk->fname1 = NULL;
	fdk->fname1_label = label;
	fdk->file_count = FIRST_FILE;

	return fdk;
}

/**
 * filedialogkeep_reset:
 * Reset FileDialogKeep.
 * Called when a user close a dialog while he has selected the first file
 * but hasn't selected the second file.
 * Input:
 * FileDialogKeep *fdk;
 * Output:
 * FileDialogKeep *fdk; updated.
 */
static void
filedialogkeep_reset(FileDialogKeep *fdk)
{
	if (fdk->fname1 == NULL)
		return;
	/* Not change: fdk->is_directory = FALSE;*/
	g_free(fdk->fname1);
	fdk->fname1 = NULL;
	gtk_widget_hide(GTK_WIDGET(fdk->fname1_label));
	fdk->file_count = FIRST_FILE;
}

/**
 * filedialogkeep_update:
 * Update FileDialogKeep.
 * Called after a user has selected the first file.
 * Input:
 * FileDialogKeep *fdk;
 * const char *fname1; Selected first file name.
 * Output:
 * FileDialogKeep *fdk; updated.
 */
static void
filedialogkeep_update(FileDialogKeep *fdk, const char *fname1)
{
	fdk->fname1 = g_strdup(fname1);
	gtk_label_set_text(fdk->fname1_label, fname1);
	gtk_widget_show(GTK_WIDGET(fdk->fname1_label));
	fdk->file_count = SECOND_FILE;
}

