/*
 * Back-end main header
 * This is the most important file for you to understand back-end.
 *
 * Programmed by INOUE Seiichiro <inoue@ainet.or.jp>
 */
#ifndef _GDIFF_GDIFF_H_
#define _GDIFF_GDIFF_H_

#ifdef __cplusplus 
extern "C" {
#endif /* __cplusplus */

/*
 * Description of data structures
 *
 * DiffDir is allocated for one session of comparing(diff) two files
 * or two directories.
 * DiffDir keeps the linked list of DiffFiles.
 * Each node(DiffFiles) is allocated for two different files under the specified directories.
 * The DiffFiles has two files information, and the linked list of DiffLines.
 * Each node(DiffLines) is allocated for different part(between lines),
 * and has the type(DiffType) and the line numbers of each different part.
 */

/* Forward declarations of types. */
typedef struct _DiffDir DiffDir;
typedef struct _DiffFiles DiffFiles;
typedef struct _FileInfo FileInfo;
typedef struct _DiffLines DiffLines;
typedef struct _DiffBeginEnd DiffBeginEnd;

/*
 * Contant number
 */
/* IMPORTANT. This is used in DiffLines structure.
 * Basically, they show diff types.
 * But, it is also used for the return value from parse_diff_line(). */
typedef enum { 
	F1ONLY,	/* Lines taken from just the first file. (known as DELETE) */
	F2ONLY,	/* Lines taken from just the second file. (known as ADD) */
	CHANGE,
	/* The followings are used for the return value from parse_diff_line() */
	FINDFILE,
	BINARY,
	IGNORE,
	ERROR,
} DiffType; 


#define NUM_COMPARE_FILES		2
typedef enum {
	FIRST_FILE = 0,
	SECOND_FILE = 1,
} WhichFile;

typedef enum {
	BINARY_FILES,
	ONLY_FILE1_EXISTS,
	ONLY_FILE2_EXISTS,
	DIFFERENT_FILES,
	IDENTICAL_FILES,
	DIRECTORIES,
} FilesSpecialStatus;

/*
 * Data structure definitions
 */

/* Directory information data
 * Note: Comparing two files is dealt with as one special case of comparing two directories.
 * - When comparing directories, the first node has the directories info,
 *   and the following nodes has each files info.
 * - When comparing two files, the first node has the files info. */
struct _DiffDir {
	GSList *dfiles_list;	/* Singly linked list of DiffFiles */
};


/* Compared files information data
 * Some special cases as follows,
 * - dlines_list==NULL implies two files are identical.
 * - One of fileinfo[]'s fname==NULL implies that the file exists only in one directory.
 *
 * struct _FileInfo
 * Two possible cases,
 * 1)directory info. 2)file info. */
struct _DiffFiles {
	struct _FileInfo {
		char *fname;	/* file name */
		int nline;		/* Number of lines */
		char *text;		/* (mmap'd)buffer */
		int lenb;		/* Length(Byte) */
		gboolean f_dir;	/* TRUE if directory */
	} fileinfo[NUM_COMPARE_FILES];
	GSList *dlines_list;/* Singly linked list of DiffLines */
	gboolean binary;
};


/* Diff lines information data
 * Keeps difftype and the line numbers. 
 * Different between these lines. */
struct _DiffLines {
	DiffType difftype;
	struct _DiffBeginEnd {
		/* the followings are line number. */
		int begin;
		int end;
	} between[NUM_COMPARE_FILES];
};


/* Global variable declarations */
/* None */

/* Global function declarations */
/* diffdata.c */
extern DiffDir* new_diff_dir(const char *fname1, const char *fname2);
extern void delete_diff_dir(DiffDir *diffdir);
extern DiffFiles* add_files_to_diff_dir(DiffDir *diffdir, const char *fname1, const char *fname2);
extern const FileInfo* get_fileinfo_from_diff_files(DiffFiles *dfiles, WhichFile n, gboolean f_mmap);
extern void add_dlines_to_diff_files(DiffFiles* dfiles, DiffType dtype, int f1n1, int f1n2, int f2n1, int f2n2);

/* rundiff.c */
extern void run_diff(DiffDir *diffdir);

/* diffmisc.c */
extern FilesSpecialStatus get_files_special_status(DiffFiles *dfiles);
extern int get_merged_numlines(DiffFiles *files);
extern int calc_diff_total_lines(const DiffFiles *dfiles, WhichFile n);
extern int get_max_nlines(DiffFiles *dfiles);

#ifdef __cplusplus
} 
#endif /* __cplusplus */
 
#endif /* __GDIFF_GDIFF_H__ */
