/*
 * GUI misc module
 * This module is for generic GUI functions.
 * Independent from specific data structure.
 *
 * Programmed by INOUE Seiichiro <inoue@ainet.or.jp>
 */
#include <stdio.h>
#include <gtk/gtk.h>
#include "gdiff.h"
#include "gui.h"
#include "misc.h"



/**
 * draw_text_lines:
 * Draw "nl" lines of buffer on GtkText widget.
 * Input:
 * GtkText *text;
 * const FontProp *fprop;
 * const char *buf; buffer.
 * int nl; the number of lines to draw.
 * int max_lenb; the length of buffer(bytes).
 * Output:
 * Return value; The drawn bytes.
 */
int
draw_text_lines(GtkText *text, const FontProp *fprop, const char *buf, int nl, int max_lenb)
{
	int len;
	const char *endp;

	endp = skip_n_lines(buf, nl, max_lenb);
	len = endp - buf;
	gtk_text_insert(text, fprop->font, fprop->fg, fprop->bg, buf, len);
	return len;
}


/**
 * insert_remove_line_numbers:
 * Insert or remove line numbers at the head of each line.
 * Input:
 * GtkText *text;
 * int *pos; Position in the text widget to insert numbers.
 * gboolean b_in; if TRUE insert numbers. if FALSE remove numbers. 
 * const FontProp *fprop;
 * const char *buf; Used to calculate the length of each line.
 * int begin; The line number to start with.
 * int nline; The number of lines to insert. (e.g. begin=10,nline=3 implies "10","11","12" are inserted.)
 * int max_lenb; Buffer length. Used to avoid an overrun.
 * int n_col; The number of columns of inserted numbers. (e.g. n_col=3 implies from "1" to "999".)
 * const char *mark; String inserted after each numbers. (e.g. "> ")
 * Output:
 * int *pos; The last position after inserting or removing numbers.
 * Return value; The length of processed part of the buffer.
 */
int insert_remove_line_numbers(GtkText *text, gboolean b_in, int *pos, const FontProp *fprop, const char *buf, int begin, int nline, int max_lenb, int n_col, const char *mark)
{
	int n;
	int end = begin + nline;
	const char *beginp = buf;
	const char *endp;
	char format[32];
	gboolean b_sbcs = !text->use_wchar;/* Single byte character set. */
	int num_char;/* number of characters */
	
	if (nline <= 0)
		return 0;

	if (b_in)
		sprintf(format, "%%%dd%s", n_col, mark);/* e.g. "%4d> " */

	/* Skip to the begin point */
	gtk_text_set_point(text, *pos);
	for (n = begin; n < end; n++) {
		int lenb;
		char buf[32];

		endp = skip_n_lines(beginp, 1, max_lenb);
		lenb = endp - beginp;
		max_lenb -= lenb;
		if (b_sbcs) {
			num_char = lenb;/* I can assume char=byte. It's a special case! */
		} else {
			num_char = get_num_chars(beginp, lenb);
		}
		beginp = endp;

		if (b_in) {
			sprintf(buf, format, n);
			gtk_text_insert(text, fprop->font, fprop->fg, fprop->bg, buf, n_col+MARK_LENGTH);
			*pos += ((n_col+MARK_LENGTH) + num_char);
		} else {
			gtk_text_forward_delete(text, n_col+MARK_LENGTH);
			*pos += num_char;
		}
		gtk_text_set_point(text, *pos);
	}
	return (endp - buf);
}


/**
 * make_label_for_notebook:
 * From two file names, make a GtkLabel widget for adding to GtkNotebook widget.
 * Input:
 * const char *f1;
 * const char *f2;
 * Output:
 * Return value; Created GtkLabel widget.
 */
GtkWidget*
make_label_for_notebook(const char *f1, const char *f2)
{
	char *buf;
	GtkWidget *label;
	
	buf = (char*)g_malloc(strlen(f1) + strlen(f2) + 4);
	sprintf(buf, "%s : %s", f1, f2);
	label = gtk_label_new(buf);
	g_free(buf);

	return label;
}
