/*
 * GUI front-end
 * GDiffDirViews, GDiffFileViews module
 * See "gui.h" for the details of data structure.
 *
 * Programmed by INOUE Seiichiro <inoue@ainet.or.jp>
 */
#include <stdio.h>
#include <gtk/gtk.h>
#include "gdiff.h"
#include "gui.h"

/* Private function declarations */
static void gdiff_fileviews_delete(GDiffFileViews *gfileviews);


/**
 * gdiff_dirviews_new:
 * Allocate GDiffDirView, intialize it and return its pointer.
 * Internally, it calls back-end routine(new_diff_dir()).
 * And it also creates widgets for displaying the result.
 * If comparing two files instead of comparing two directories,
 * it internally allocates GDiffFileViews and shows its result.
 * But the GDiffFileViews will be finalized via gdiff_dirviews_delete().
 * Input:
 * GDiffWindow *gdwin; kept in GDiffDirView for reference.
 * const char *fname1; first file name.
 * const char *fname2; second file name.
 * gboolean isdir; TRUE if they are directories.
 * Output:
 * GDiffWindow *gdwin; GList *gdirv_list is updated.
 * Return value; Allocated GDiffDirViews.
 */
GDiffDirViews*
gdiff_dirviews_new(GDiffWindow *gdwin, const char *fname1, const char *fname2, gboolean isdir)
{
	GDiffDirViews *gdirviews;

	gdirviews = g_new(GDiffDirViews, 1);

	/* Back-end data */
	gdirviews->diffdir = new_diff_dir(fname1, fname2);
	gdirviews->isdir = isdir;
	gdirviews->gdwin = gdwin;
	gdirviews->gdfile_list = NULL;

	/* Can be refererd to from GDiffWindow. */
	gdwin->gdirv_list = g_list_append(gdwin->gdirv_list, gdirviews);
	
	if (isdir) {
		gdirviews->dirname1 = g_strdup(fname1);
		gdirviews->dirname2 = g_strdup(fname2);

		/* Create widgets for display file list */
		create_dir_widgets(gdirviews);
	} else {
		/* compare two files */
		DiffFiles *dfiles;
		GDiffFileViews *gfileviews;

		gdirviews->dirname1 = NULL;
		gdirviews->dirname2 = NULL;
		gdirviews->base = NULL;
		gdirviews->hbox = NULL;
		gdirviews->clist = NULL;
		dfiles = g_slist_nth_data(gdirviews->diffdir->dfiles_list, 0);
		gfileviews = gdiff_fileviews_new(gdwin, gdirviews, dfiles, g_pref.pane_mode);
	}

	return gdirviews;
}

/**
 * gdiff_dirviews_delete:
 * Finalize GDiffDirViews structure, and free its memory.
 * It also destroys the related widgets.
 * Every related GDiffFileViews and their related widgets will be finalized, too.
 * Input:
 * GDiffDirViews *gdirviews;
 * Output:
 * None;
 */
void
gdiff_dirviews_delete(GDiffDirViews *gdirviews)
{
	GSList *list;/* Node of GDiffFileViews */

	g_return_if_fail(gdirviews != NULL);

	g_print("dirview delete; len=%d\n", g_slist_length(gdirviews->gdfile_list));
	for (list = gdirviews->gdfile_list; list; list = list->next) {
		gdiff_fileviews_delete(list->data);
	}
	g_slist_free(gdirviews->gdfile_list);
	
	/* Front-end data */
	if (gdirviews->base) {
		gint page;
		page = gtk_notebook_page_num(gdirviews->gdwin->notebook, GTK_WIDGET(gdirviews->base));
		/*gtk_notebook_remove_page(gdirviews->gdwin->notebook, page);*/
		gtk_widget_destroy(GTK_WIDGET(gdirviews->base));
	}

	/* Back-end data */
	delete_diff_dir(gdirviews->diffdir);
	if (gdirviews->dirname1)
		g_free(gdirviews->dirname1);
	if (gdirviews->dirname2)
		g_free(gdirviews->dirname2);

	/* Free its own memory */
	g_free(gdirviews);
}


/**
 * gdiff_current_views:
 * Search the current view in the notebook widget,
 * and return its related data strucure(GDiffDirViews or GDiffFileViews).
 * This doesn't return a proper type.
 * It is up to the caller to cast it to the proper type depending on ret_isdir.
 * Input:
 * const GDiffWindow *gdwin;
 * Output:
 * gboolean *ret_isdir; return TRUE if it is directory.
 * Return value; pointer to GDiffDirViews or GDiffFileViews. In no view case,  NULL.
 */
void*
gdiff_current_views(const GDiffWindow *gdwin, gboolean *ret_isdir)
{
	GList *list;
	GtkNotebook *notebook = gdwin->notebook;
	GtkWidget *w;
	gint page;

	page = gtk_notebook_current_page(notebook);
	if (page < 0)
		return NULL;
	w = gtk_notebook_get_nth_page(notebook, page);

	for (list = gdwin->gdirv_list; list; list = list->next) {
		GDiffDirViews *gdviews = (GDiffDirViews*)list->data;
		if (gdviews->base == w) {
			*ret_isdir = TRUE;
			return gdviews;
		}
	}
	for (list = gdwin->gfilev_list; list; list = list->next) {
		GDiffFileViews *gfviews = (GDiffFileViews*)list->data;
		if (gfviews->base == w) {
			*ret_isdir = FALSE;
			return gfviews;
		}
	}
	
	return NULL;
}


/**
 * gdiff_fileviews_new:
 * Allocate GDiffFileViews, intialize it and return its pointer.
 * And it also creates widgets for displaying the result.
 * Input:
 * GDiffWindow *gdwin; kept in GDiffFileViews for reference.
 * GDiffDirViews *gdirviews; kept in GDiffFileViews for reference.
 * const DiffFiles *dfiles; (Back-end data)kept in GDiffFileViews for reference.
 * PaneMode pmode; ONE_PANE or TWO_PANE.
 * Ouput:
 * Return value; Allocated GDiffFileViews.
 */
GDiffFileViews*
gdiff_fileviews_new(GDiffWindow *gdwin, GDiffDirViews *gdirviews, DiffFiles *dfiles, PaneMode pmode)
{
	GDiffFileViews *gfileviews;
	const FileInfo *fi1;
	const FileInfo *fi2;

	gfileviews = g_new(GDiffFileViews, 1);
	gfileviews->dfiles = dfiles;

	/* Can be referrerd to from each other. */
	gdirviews->gdfile_list = g_slist_append(gdirviews->gdfile_list, gfileviews);
	gfileviews->gdirviews = gdirviews;

	/* Can be referrerd to from GDiffWindow. */
	gdwin->gfilev_list = g_list_append(gdwin->gfilev_list, gfileviews);
	gfileviews->gdwin = gdwin;

	fi1 = get_fileinfo_from_diff_files(dfiles, FIRST_FILE, FALSE);
	fi2 = get_fileinfo_from_diff_files(dfiles, SECOND_FILE, FALSE);
	gfileviews->filename1 = g_strdup(fi1->fname);
	gfileviews->filename2 = g_strdup(fi2->fname);

	gfileviews->state.pane_mode = pmode;
	
	/* Create widgets for display diff files */
	if (gfileviews->state.pane_mode == ONE_PANE)
		create_onepane_widgets(gfileviews);
	else if (gfileviews->state.pane_mode == TWO_PANE)
		create_twopane_widgets(gfileviews);

	gfileviews->merged_nline = 0;
			
	return gfileviews;
}


/**
 * gdiff_fileviews_delete:
 * Finalize GDiffFileViews structure, and free its memory.
 * It also destroys the related widgets.
 * Input:
 * GDiffFileViews *gfileviews;
 * Output:
 * None;
 */
static void
gdiff_fileviews_delete(GDiffFileViews *gfileviews)
{
	g_return_if_fail(gfileviews != NULL);

	g_print("really destroy\n");
	g_free(gfileviews->filename1);
	g_free(gfileviews->filename2);

	/* XXX: to destory is enough ? */
	/* Remove its own page from the notebook widget */
	if (gfileviews->base) {
		gint page;
		page = gtk_notebook_page_num(gfileviews->gdwin->notebook, GTK_WIDGET(gfileviews->base));
		/*gtk_notebook_remove_page(gfileviews->gdwin->notebook, page);*/
		gtk_widget_destroy(GTK_WIDGET(gfileviews->base));
	}

	/* Free its own memory */
	g_free(gfileviews);
}


