/*
 * Misc module
 * This module is for generic functions.
 * Independent from specific data structure.
 *
 * Programmed by INOUE Seiichiro <inoue@ainet.or.jp>
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#if defined(HAVE_STRING_H)
#include <string.h>
#elif defined(HAVE_STRINGS_H)
#include <strings.h>
#endif
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <glib.h>
#include <gdk/gdk.h>/* for gdk_mbstowcs */
#include "misc.h"


/**
 * skip_n_lines:
 * Skip "nl" lines in the buffer, and return the head of the line.
 * Input:
 * const char *ptr;
 * int nl; Number of lines to skip
 * int max_lenb; buffer length(byte)
 * Output:
 * Return value; Pointer to the head of the line.
 */
const char*
skip_n_lines(const char *ptr, int nl, int max_lenb)
{
	int i;
	int ml = 0;

	if (nl <= 0)
		return ptr;
	for (i = 0; i < max_lenb; i++) {
		if (*ptr++ == '\n') {
			ml++;
			if (ml == nl)
				break;
		}
	}
	return ptr;
}


/**
 * check_filetype:
 * Simple and generic file type check.
 * Input:
 * const char *fname; file name.
 * Output:
 * Return value: FileType;
 */
FileType
check_filetype(const char *fname)
{
	struct stat sb;

	if (stat(fname, &sb) == -1) {
		g_warning("%s: fstat in is_dir()", fname);
		return OTHERFILE;
	}
	if (sb.st_mode & S_IFREG)
		return REGULARFILE;
	else if (sb.st_mode & S_IFDIR)
		return DIRECTORY;
	else
		return OTHERFILE;
}


/**
 * get_num_chars:
 * Retrun the number of characters in the buffer.
 * In SBCS(Single Byte Character Set) environment, not called for efficiency.
 * In that case, you can assume the number of characters is the same as one of bytes.
 * Input:
 * const char *src; buffer, normally MBCS.
 * int lenb; length in bytes.
 * Output:
 * Return value; the number of characters.
 */
int
get_num_chars(const char *src, int lenb)
{
	int num_char;
	GdkWChar *pwcs;
	GdkWChar wcs[BUFSIZ];
	char *pmbs;
	char mbs[BUFSIZ];/* buffer to keep nul-terminated string */
	gboolean buf_allocated = FALSE;

	if (lenb < BUFSIZ) {
		pwcs = wcs;
		pmbs = mbs;
	} else {
		pmbs = (char*)g_malloc(lenb+1);
		pwcs = (GdkWChar*)g_malloc((lenb+1)*sizeof(GdkWChar));
		buf_allocated = TRUE;
	}
	memcpy(pmbs, src, lenb);
	pmbs[lenb] = '\0';
	/*XXX:this interface is better.
	  num_char = gdk_mbstowcs(NULL, pbuf, 0);*/
	num_char = gdk_mbstowcs(pwcs, pmbs, lenb+1);
	
	if (buf_allocated == TRUE) {
		g_free(pmbs);
		g_free(pwcs);
	}	
	return num_char;
}

/**
 * calc_number_places:
 * Calculate the number of places on base 10.
 * e.g. 789 => 3, 7890 => 4, 78901 => 5 ...
 * Input:
 * int n;
 * Output:
 * Return value;
 */
int
calc_number_places(int n)
{
	int ret  = 0;

	while (n) {
		n /= 10;
		ret++;
	}
	return ret;
}

/**
 * close_allfd_with_except:
 * close almost all file descriptors.
 * The following descriptors are not closed.
 * stdin, stdout, stderr, and the fd's in except[] array.
 * Usually, called by forked child process,
 * and its main intension is to close fd of socket connected to X server.
 * Input:
 * const int except[]; fd's not to be closed.
 * int num_ex; number of fd's in except array.
 * Output:
 * None;
 */
void
close_allfd_with_except(const int except[], int num_ex)
{
	int fd_max;
	int fd;

	fd_max = sysconf(_SC_OPEN_MAX);
	for (fd = STDERR_FILENO+1; fd < fd_max; fd++) {
		int ne;
		gboolean b_close = TRUE;
		for (ne = 0; ne < num_ex; ne++) {
			if (fd == except[ne])
				b_close = FALSE;
		}
		if (b_close)
			close(fd);
	}
}

/**
 * def_signal_handler:
 * default (Unix)signal handler.
 * Now, it takes care of child process death.
 */
void
def_signal_handler(int sig)
{
	int sta;

	if (sig == SIGCHLD)
		wait(&sta);
}
