/*
 * Module for file view(display "diff" result) in one-pane mode
 *
 * Programmed by INOUE Seiichiro <inoue@ainet.or.jp>
 */
#include <stdio.h>
#include <gtk/gtk.h>
#include "gdiff.h"
#include "gui.h"
#include "misc.h"


/* Private function declarations */
static void init_state(GDiffFileViews *gfileviews);
static void create_textw(GDiffFileViews *gfileviews);
static void draw_text(GDiffFileViews *gfileviews);
static void show_hide_numbers(GDiffFileViews *gfileviews, gboolean b_show);
static void create_overview(GDiffFileViews *gfileviews);
static void draw_overview(GDiffFileViews *gfileviews);
static void draw_overview_lines(GtkWidget *widget, GdkEventExpose *event, gpointer data);


/**
 * create_onepane_widgets:
 * Called from gdiff_fileviews_new().
 * Create widgets and show them.
 * The contents(diff results) are not shown here, which is up to one_pane_display().
 * Input:
 * GDiffFileViews *gfileviews;
 * Output:
 * GDiffFileViews *gfileviews; updated.
 */
void
create_onepane_widgets(GDiffFileViews *gfileviews)
{
	GDiffWindow *gdwin = gfileviews->gdwin;
	GtkWidget *hbox;
	GtkWidget *label;
	gint page;
	int i;
	
	/* hbox */
	hbox = gtk_hbox_new(FALSE, 0);
	gfileviews->pane.one.hbox = GTK_HBOX(hbox);
	gfileviews->base = GTK_WIDGET(gfileviews->pane.one.hbox);/* Virtual base widget. */

	/* Add to the notebook */
	label = make_label_for_notebook(gfileviews->filename1, gfileviews->filename2);
	gfileviews->pane.one.label = GTK_LABEL(label);
	gtk_notebook_append_page(gdwin->notebook, hbox, label);

	/* text area */
	create_textw(gfileviews);
	/* overview area */
	create_overview(gfileviews);
	
	gtk_widget_show(hbox);
	/* Is this a proper way ? */
	page = gtk_notebook_page_num(gdwin->notebook, hbox);
	gtk_notebook_set_page(gdwin->notebook, page);

	create_rightclick_menu(gfileviews);
	
	/*XXX: color */
	/* Now, color managing is a kind of kludge.
	   Because the color of text and overview should be synchronized,
	   I'm getting color from overview and use this for text. */
	for (i = 0; i < NUM_COMPARE_FILES; i++) {
		gfileviews->fcolor[i] = &(GTK_WIDGET(gfileviews->pane.one.overview[i])->style->fg[GTK_STATE_NORMAL]);
	}
}

/**
 * one_pane_display:
 * Show the diff result in one-pane mode.
 * Input:
 * GDiffFileViews *gfileviews;
 * Output:
 * GDiffFileViews *gfileviews; updated.
 */
void
one_pane_display(GDiffFileViews *gfileviews)
{
	DiffFiles *files = gfileviews->dfiles;
	GtkText *text;
	FilesSpecialStatus fsst;

	fsst = get_files_special_status(files);
	if (fsst != DIFFERENT_FILES) {
		g_message("binary or identical or only one file exists.(%d)\n", fsst);
		return;
	}
	init_state(gfileviews);
	draw_text(gfileviews);
	draw_overview(gfileviews);
	if (gfileviews->state.show_line_num == TRUE) {
		show_hide_numbers(gfileviews, TRUE);
	}

	/* Set right-click menu */
	text = gfileviews->pane.one.text;
	gtk_signal_connect(GTK_OBJECT(text), "button_press_event",
					   GTK_SIGNAL_FUNC(rightclick_menu_cb), gfileviews);
}

/**
 * one_pane_show_linenum:
 * Show(Hide) the line numbers on text widget.
 * Input:
 * GDiffFileViews *gfileviews;
 * gboolean to_show; If TRUE, show. if FALSE, hide.
 * Output:
 * GDiffFileViews *gfileviews; updated.
 */
void
one_pane_show_linenum(GDiffFileViews *gfileviews, gboolean to_show)
{
	if (to_show == TRUE && gfileviews->state.show_line_num == FALSE) {
		show_hide_numbers(gfileviews, TRUE);
		gfileviews->state.show_line_num = TRUE;
	} else if (to_show == FALSE && gfileviews->state.show_line_num == TRUE) {
		show_hide_numbers(gfileviews, FALSE);
		gfileviews->state.show_line_num = FALSE;
	}	
}

#if 0
/**
 * one_pane_next_prev:
 */ 
void
one_pane_next_prev(GDiffFileViews *gfileviews, gboolean to_next)
{
	if (to_next == TRUE)
		move_next(gfileviews);
	else
		move_prev(gfileviews);
}
#endif


/* ---The followings are private functions--- */
static void init_state(GDiffFileViews *gfileviews)
{
	gfileviews->state.show_line_num = g_pref.show_line_num;
}

/**
 * create_textw:
 * Create GtkText widget.
 */
static void
create_textw(GDiffFileViews *gfileviews)
{
	GtkHBox *hbox = gfileviews->pane.one.hbox;
	GtkWidget *text;
	GtkWidget *scrolled_window;

	scrolled_window = gtk_scrolled_window_new(NULL, NULL);
	gtk_box_pack_start(GTK_BOX(hbox), scrolled_window, TRUE, TRUE, 0);
	gtk_widget_show(scrolled_window);
	
	text = gtk_text_new(NULL, NULL);
	gtk_text_set_editable(GTK_TEXT(text), FALSE);
	gtk_text_set_word_wrap(GTK_TEXT(text), FALSE);
	gtk_text_set_line_wrap(GTK_TEXT(text), g_pref.line_wrap);
	gtk_container_add(GTK_CONTAINER(scrolled_window), text);
	gtk_widget_grab_focus(text);
	gtk_widget_show(text);

	gfileviews->pane.one.text = GTK_TEXT(text);
}


static void
draw_text(GDiffFileViews *gfileviews)
{
	DiffFiles *files = gfileviews->dfiles;
	const FileInfo *fi1 = get_fileinfo_from_diff_files(files, FIRST_FILE, TRUE);
	const FileInfo *fi2 = get_fileinfo_from_diff_files(files, SECOND_FILE, TRUE);
	GtkText *text = gfileviews->pane.one.text;
	GdkFont *font = GTK_WIDGET(text)->style->font;
	const char *pt1;
	const char *pt2;
	const char *text_last1;/* for calculation of the length left */
	const char *text_last2;/* for calculation of the length left */
	int line1;
	GSList *node;/* node of DiffLines' list */
	FontProp fprop;
	GdkColor fcolor_common = {0, 0xFFFF, 0xFFFF, 0xFFFF};/* white */
	GdkColor *fcolor_f1 = gfileviews->fcolor[FIRST_FILE];
	GdkColor *fcolor_f2 = gfileviews->fcolor[SECOND_FILE];

	fprop.font = font;
	fprop.fg = NULL;

	line1 = 1;
	pt1 = fi1->text;
	text_last1 = fi1->text + fi1->lenb;
	text_last2 = fi2->text + fi2->lenb;
	gtk_text_freeze(text);
	for (node = files->dlines_list; node; node = node->next) {
		DiffLines *dlines;
		int begin1, end1;
		int begin2, end2;
		int lenb;
		
		dlines = (DiffLines*)(node->data);
		/* Use local variables for ease to read */
		begin1 = dlines->between[FIRST_FILE].begin;
		end1 = dlines->between[FIRST_FILE].end;
		begin2 = dlines->between[SECOND_FILE].begin;
		end2 = dlines->between[SECOND_FILE].end;

		if (end1 != 0) {
			fprop.bg = &fcolor_common;
			lenb = draw_text_lines(text, &fprop, pt1, begin1 - line1, text_last1 - pt1);
			pt1 += lenb;
		}
		switch (dlines->difftype) {
		case CHANGE:
			fprop.bg = fcolor_f1;
			lenb = draw_text_lines(text, &fprop, pt1, end1-begin1+1, text_last1 - pt1);
			pt1 += lenb;

			fprop.bg = fcolor_f2;
			pt2 = skip_n_lines(fi2->text, begin2-1, fi2->lenb);/*XXX*/
			draw_text_lines(text, &fprop, pt2, end2-begin2+1, text_last2 - pt2);
			break;
		case F2ONLY:
			fprop.bg = &fcolor_common;
			if (end1 != 0) {
				lenb = draw_text_lines(text, &fprop, pt1, end1-begin1+1, text_last1 - pt1);
				pt1 += lenb;
			}
			fprop.bg = fcolor_f2;
			pt2 = skip_n_lines(fi2->text, begin2-1, fi2->lenb);/*XXX*/
			draw_text_lines(text, &fprop, pt2, end2-begin2+1, text_last2 - pt2);
			break;
		case F1ONLY:
			fprop.bg = fcolor_f1;
			lenb = draw_text_lines(text, &fprop, pt1, end1-begin1+1, text_last1 - pt1);
			pt1 += lenb;
			break;
		default:/* not used */
			break;
		}
		line1 = end1 + 1;
	}
	/* Draw the remained part */
	if (line1 >= fi1->nline) {
		g_print("Problem?\n");
	}
	fprop.bg = &fcolor_common;
	draw_text_lines(text, &fprop, pt1, fi1->nline - line1 + 1, text_last1 - pt1);
	gtk_text_thaw(text);
}


/* line numbers */
static void
show_hide_numbers(GDiffFileViews *gfileviews, gboolean b_show)
{
	DiffFiles *files = gfileviews->dfiles;
	const FileInfo *fi1 = get_fileinfo_from_diff_files(files, FIRST_FILE, TRUE);
	const FileInfo *fi2 = get_fileinfo_from_diff_files(files, SECOND_FILE, TRUE);
	GtkText *text = gfileviews->pane.one.text;
	GdkFont *font = GTK_WIDGET(text)->style->font;
	const char *pt1;
	const char *pt2;
	const char *text_last1;/* for calculation of the length left */
	const char *text_last2;/* for calculation of the length left */
	int line1;
	int n_col;/* the number of columns for representing line numbers */
	int pos = 0;/* position in text widget */
	int merged_nline = gfileviews->merged_nline;
	GSList *node;/* node of DiffLines' list */
	FontProp fprop;
	GdkColor fcolor_common = {0, 0xFFFF, 0xFFFF, 0xFFFF};/* white */
	GdkColor *fcolor_f1 = gfileviews->fcolor[FIRST_FILE];
	GdkColor *fcolor_f2 = gfileviews->fcolor[SECOND_FILE];

	fprop.font = font;
	fprop.fg = NULL;

	line1 = 1;
	pt1 = fi1->text;
	text_last1 = fi1->text + fi1->lenb;
	text_last2 = fi2->text + fi2->lenb;
	n_col = calc_number_places(merged_nline);
	gtk_text_freeze(text);
	for (node = files->dlines_list; node; node = node->next) {
		DiffLines *dlines;
		int begin1, end1;
		int begin2, end2;
		int lenb;
		
		dlines = (DiffLines*)(node->data);
		/* Use local variables for ease to read */
		begin1 = dlines->between[FIRST_FILE].begin;
		end1 = dlines->between[FIRST_FILE].end;
		begin2 = dlines->between[SECOND_FILE].begin;
		end2 = dlines->between[SECOND_FILE].end;

		if (end1 != 0) {
			fprop.bg = &fcolor_common;
			lenb = insert_remove_line_numbers(text, b_show, &pos,
											  &fprop, pt1, 
											  line1, begin1 - line1, text_last1 - pt1,
											  n_col, MARK_COMMON);
			pt1 += lenb;
		}
		switch (dlines->difftype) {
		case CHANGE:
			fprop.bg = fcolor_f1;
			lenb = insert_remove_line_numbers(text, b_show, &pos,
											  &fprop, pt1,
											  begin1, end1 - begin1 + 1, text_last1 - pt1,
											  n_col, MARK_FILE1);
			pt1 += lenb;

			fprop.bg = fcolor_f2;
			pt2 = skip_n_lines(fi2->text, begin2-1, fi2->lenb);
			lenb = insert_remove_line_numbers(text, b_show, &pos,
											  &fprop, pt2,
											  begin2, end2 - begin2 + 1, text_last2 - pt2,
											  n_col, MARK_FILE2);
			break;
		case F2ONLY:
			fprop.bg = &fcolor_common;
			if (end1 != 0) {
				lenb = insert_remove_line_numbers(text, b_show, &pos,
												  &fprop, pt1,
												  begin1, end1 - begin1 + 1, text_last1 - pt1,
												  n_col, MARK_COMMON);
				pt1 += lenb;
			}

			fprop.bg = fcolor_f2;
			pt2 = skip_n_lines(fi2->text, begin2-1, fi2->lenb);
			lenb = insert_remove_line_numbers(text, b_show, &pos,
											  &fprop, pt2,
											  begin2, end2 - begin2 + 1, text_last2 - pt2,
											  n_col, MARK_FILE2);
			break;
		case F1ONLY:
			fprop.bg = fcolor_f1;
			lenb = insert_remove_line_numbers(text, b_show, &pos,
											  &fprop, pt1,
											  begin1, end1 - begin1 + 1, text_last1 - pt1,
											  n_col, MARK_FILE1);
			pt1 += lenb;
			break;
		default:/* not used */
			break;
		}
		line1 = end1 + 1;
	}
	/* Draw the remained part */
	fprop.bg = &fcolor_common;
	insert_remove_line_numbers(text, b_show, &pos, &fprop, pt1,
						line1, fi1->nline - line1 + 1, text_last1 - pt1,
						n_col, MARK_COMMON);
	gtk_text_thaw(text);
}


/* overview */
/**
 * create_overview:
 * Create GdiffOverview widgets. See "gdiffoverview.[ch]".
 */
static void
create_overview(GDiffFileViews *gfileviews)
{
	GtkHBox *hbox = gfileviews->pane.one.hbox;
	GtkAdjustment *vadj;
	GtkWidget *ov1;
	GtkWidget *ov2;
	GtkWidget *darea;/* drawing area */

	
	darea = gtk_drawing_area_new();
	gtk_widget_set_name(GTK_WIDGET(darea), "drawing overview");
	vadj = GTK_TEXT(gfileviews->pane.one.text)->vadj;
	ov1 = gdiff_overview_new(vadj);
	gtk_widget_set_name(GTK_WIDGET(ov1), "file1 overview");
	ov2 = gdiff_overview_new(vadj);
	gtk_widget_set_name(GTK_WIDGET(ov2), "file2 overview");

	gtk_box_pack_start(GTK_BOX(hbox), ov1, FALSE, FALSE, 0);
	gtk_widget_show(ov1);
	gtk_box_pack_start(GTK_BOX(hbox), darea, FALSE, FALSE, 0);
	gtk_widget_show(darea);
	gtk_box_pack_start(GTK_BOX(hbox), ov2, FALSE, FALSE, 0);
	gtk_widget_show(ov2);
	
	/* adjust drawing-area */
	gtk_drawing_area_size(GTK_DRAWING_AREA(darea), WIDTH_OVERVIEW, 0);/*XXX*/
	gdiff_overview_size(GDIFF_OVERVIEW(ov1), WIDTH_OVERVIEW, 0);/*XXX*/
	gdiff_overview_size(GDIFF_OVERVIEW(ov2), WIDTH_OVERVIEW, 0);/*XXX*/
	
	gfileviews->pane.one.darea = GTK_DRAWING_AREA(darea);
	gfileviews->pane.one.overview[FIRST_FILE] = GDIFF_OVERVIEW(ov1);
	gfileviews->pane.one.overview[SECOND_FILE] = GDIFF_OVERVIEW(ov2);
}

static void
draw_overview(GDiffFileViews *gfileviews)
{
	DiffFiles *files = gfileviews->dfiles;
	GdiffOverview *ov1 = gfileviews->pane.one.overview[FIRST_FILE];
	GdiffOverview *ov2 = gfileviews->pane.one.overview[SECOND_FILE];
	GtkDrawingArea *darea = gfileviews->pane.one.darea;
	GSList *node;/* node of DiffLines' list */
	int merged_nline;
	int drawn_nlines1 = 0;
	int drawn_nlines2 = 0;
	
	gfileviews->merged_nline = merged_nline = get_merged_numlines(files);
	if (merged_nline == 0) {
		g_warning("impossible: merged_nline == 0\n");
		return;
	}
	
	for (node = files->dlines_list; node; node = node->next) {
		DiffLines *dlines;
		int begin1, end1;
		int begin2, end2;
		int draw_nline1;
		int draw_nline2;
		int line;
		
		dlines = (DiffLines*)(node->data);
		/* Use local variables for ease to read */
		begin1 = dlines->between[FIRST_FILE].begin;
		end1 = dlines->between[FIRST_FILE].end;
		begin2 = dlines->between[SECOND_FILE].begin;
		end2 = dlines->between[SECOND_FILE].end;

		draw_nline1 = end1 - begin1 + 1;
		line = begin1 + drawn_nlines2;
		gdiff_overview_insert_paintrange(ov1, (gdouble)line/merged_nline, (gdouble)(line+draw_nline1)/merged_nline);

		draw_nline2 = end2 - begin2 + 1;
		line = begin2 + drawn_nlines1;
		gdiff_overview_insert_paintrange(ov2, (gdouble)line/merged_nline, (gdouble)(line+draw_nline2)/merged_nline);

		switch (dlines->difftype) {
		case CHANGE:
			drawn_nlines1 += draw_nline1;
			drawn_nlines2 += draw_nline2;
			break;
		case F1ONLY:
			drawn_nlines1 += draw_nline1;
			break;
		case F2ONLY:
			drawn_nlines2 += draw_nline2;
			break;
		default:/* not used */
			break;
		}
	}
	/* Overview drawing area */
	gtk_signal_connect(GTK_OBJECT(darea), "expose_event",
                      (GtkSignalFunc)draw_overview_lines, gfileviews);
	
	return;
}

static void
draw_overview_lines(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
	GDiffFileViews *gfileviews = (GDiffFileViews*)data;
	const DiffFiles *files = gfileviews->dfiles;
	GSList *node;/* node of DiffLines' list */
	GdkWindow *w = widget->window;
	GdkGC *gc = widget->style->fg_gc[GTK_STATE_NORMAL];
	int darea_width;
	int darea_height;
	int merged_nline = gfileviews->merged_nline;
	int drawn_nlines1 = 0;
	int drawn_nlines2 = 0;

	gdk_window_get_size(w, &darea_width, &darea_height);
	for (node = files->dlines_list; node; node = node->next) {
		int y1, y2;
		DiffLines *dlines;
		int begin1, end1;
		int begin2, end2;
		int draw_nline1;
		int draw_nline2;
		int line;

		dlines = (DiffLines*)(node->data);
		/* Use local variables for ease to read */
		begin1 = dlines->between[FIRST_FILE].begin;
		end1 = dlines->between[FIRST_FILE].end;
		begin2 = dlines->between[SECOND_FILE].begin;
		end2 = dlines->between[SECOND_FILE].end;

		line = begin1 + drawn_nlines2;
		draw_nline1 = end1 - begin1 + 1;
		y1 = darea_height * ((double)(line+(line+draw_nline1))/(2*merged_nline));
		line = begin2 + drawn_nlines1;
		draw_nline2 = end2 - begin2 + 1;
		y2 = darea_height * ((double)(line+(line+draw_nline2))/(2*merged_nline));
		gdk_draw_line(w, gc, 0, y1, darea_width, y2);

		switch (dlines->difftype) {
		case CHANGE:
			drawn_nlines1 += draw_nline1;
			drawn_nlines2 += draw_nline2;
			break;
		case F1ONLY:
			drawn_nlines1 += draw_nline1;
			break;
		case F2ONLY:
			drawn_nlines2 += draw_nline2;
			break;
		default:/* not used */
			break;
		}
	}
}
