/*
 * Back-end sub module
 * Execute "diff" command, and from this output create backend-data structure.
 * See "gdiff.h" for the details of data structure.
 * This module should be independent from GUI frontend.
 *
 * Programmed by INOUE Seiichiro <inoue@ainet.or.jp>
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#if defined(HAVE_STRING_H)
#include <string.h>
#elif defined(HAVE_STRINGS_H)
#include <strings.h>
#endif
#include <glib.h>
#include "gdiff.h"
#include "misc.h"


/* Constant strings */
#ifndef DIFF_PATH	/* normally defined in config.h */
#	define DIFF_PATH  "/usr/bin/diff"
#endif
#define DIFF_ARGS		"-r"


/* Private function declarations */
static DiffType parse_diff_line(const char *buf, int *f1n1, int *f1n2, int *f2n1, int *f2n2);
static void parse_files(const char *buf, char *fname1, char *fname2, const char *dir1, const char *dir2);

/**
 * run_diff:
 * Create back-end data structure from "diff" output.
 * See "gdiff.h" for the details of data structures.
 * Input:
 * DiffDir *diffdir; The first node of GSList *dfiles_list keeps the target files.
 * Output:
 * DiffDir *diffdir; GSList *dfiles_list is updated.
 */
void
run_diff(DiffDir *diffdir)
{
	DiffFiles *first_files;
	DiffFiles *current_files;
	FILE *fpdiff;
	const FileInfo *fi1;
	const FileInfo *fi2;
	char buf[BUFSIZ];
	char *prog = DIFF_PATH;
	char *args = DIFF_ARGS;
	DiffType dtype;

	first_files = current_files = g_slist_nth_data(diffdir->dfiles_list, 0);
	fi1 = get_fileinfo_from_diff_files(first_files, FIRST_FILE, FALSE);
	fi2 = get_fileinfo_from_diff_files(first_files, SECOND_FILE, FALSE);
	if (fi1->f_dir == TRUE) {
		g_print("directory compare in run_diff\n");
	}
	
	fpdiff = spawn_diff(prog, args, fi1->fname, fi2->fname);

	while (fgets(buf, sizeof(buf), fpdiff) != NULL) {
		int f1n1, f1n2, f2n1, f2n2;
		char fname1[PATH_MAX+1];
		char fname2[PATH_MAX+1];

		dtype = parse_diff_line(buf, &f1n1, &f1n2, &f2n1, &f2n2);
		switch (dtype) {
		case FINDFILE:
			parse_files(buf, fname1, fname2, fi1->fname, fi2->fname);
			current_files = add_files_to_diff_dir(diffdir, fname1, fname2);
			break;
		case F2ONLY:
		case CHANGE:
		case F1ONLY:
			add_dlines_to_diff_files(current_files, dtype, f1n1, f1n2, f2n1, f2n2);
			break;

		case BINARY:/* kludge */
			current_files->binary = TRUE;
			break;
		case IGNORE:
			break;
		case ERROR:
			g_warning("diff error?\n %s", buf);
			break;
		}
	}
}

  

/* ---The followings are private functions--- */
/* Derived from mgdiff, and modified by INOUE. */
/**
 * parse_diff_line:
 * Input:
 * const char *buf; diff output.
 * Output:
 * int *f1n1, *f1n2; First file diff part, between line f1n1 to f1n2.
 * int *f2n1, *f2n2; Second file diff part, between line f2n1 to f2n2.
 */
/* 
 * this code taken from "ediff.c" by David MacKenzie, a published,
 * uncopyrighted program to translate diff output into plain English 
 */
static DiffType
parse_diff_line(const char *buf, int *f1n1, int *f1n2, int *f2n1, int *f2n2)
{
    if ((buf[0] == '<') || (buf[0] == '>') || (buf[0] == '-')) {
		return IGNORE;
    } else if (strncmp(buf, "diff", 4) == 0 || strncmp(buf, "Only", 4) == 0) {
		return FINDFILE;
	} else if (strncmp(buf, "Binary", 6) == 0) {
		return BINARY;
    } else if (sscanf (buf, "%d,%dc%d,%d\n", f1n1, f1n2, f2n1, f2n2) == 4) {
		return CHANGE;
    } else if (sscanf (buf, "%d,%dc%d\n", f1n1, f1n2, f2n1) == 3) {
		*f2n2 = *f2n1;
		return CHANGE;
    } else if (sscanf (buf, "%dc%d,%d\n", f1n1, f2n1, f2n2) == 3) {
		*f1n2 = *f1n1;
        return CHANGE;
    } else if (sscanf (buf, "%dc%d\n", f1n1, f2n1) == 2) {
        *f2n2 = *f2n1;
        *f1n2 = *f1n1;
        return CHANGE;
    } else if (sscanf (buf, "%d,%dd%d\n", f1n1, f1n2, f2n1) == 3) {
		*f2n2 = *f2n1;
		return F1ONLY;
    } else if (sscanf (buf, "%dd%d\n", f1n1, f2n1) == 2) {
		*f2n2 = *f2n1;
		*f1n2 = *f1n1;
		return F1ONLY;
    } else if (sscanf (buf, "%da%d,%d\n", f1n1, f2n1, f2n2) == 3) {
		*f1n2 = *f1n1;
		return F2ONLY;
    } else if (sscanf (buf, "%da%d\n", f1n1, f2n1) == 2) {
		*f1n2 = *f1n1;
		*f2n2 = *f2n1;
		return F2ONLY;
    } else
		return ERROR;
}


/**
 * parse_files:
 * Parse the string as following format,
 *  - diff -r file-name1 file-name2
 *  - Only in dir-name: file-name1
 * Input:
 * const char *buf; Parsed string.
 * const char *dir1; Used for parsing the path including directory name.
 * const char *dir2; Used for parsing the path including directory name.
 * Output:
 * char *fname1; First file name. No file case, it will be '\0'.
 * char *fname2; Second file name. No file case, it will be '\0'.
 */
static void
parse_files(const char *buf, char *fname1, char *fname2, const char *dir1, const char *dir2)
{
	char tmp[PATH_MAX+1];
	char tmpdir[PATH_MAX+1];

	if (sscanf(buf, "diff -r %s %s\n", fname1, fname2) == 2)
		return;
	if (sscanf(buf, "Only in %s %s\n", tmpdir, tmp) == 2) {
		int len;

		len = strlen(tmpdir);
		/* remove the last ':' and make the last char '/' */
		if (tmpdir[len-2] != '/')
			tmpdir[len-1] = '/';
		else
			tmpdir[len-1] = '\0';
		len = strlen(dir1);
		if (strncmp(dir1, tmpdir, len) == 0) {
			strcpy(fname1, tmpdir);
			strcat(fname1, tmp);
			fname2[0] = '\0';
			return;
		}
		len = strlen(dir2);
		if (strncmp(dir2, tmpdir, len) == 0) {
			strcpy(fname2, tmpdir);
			strcat(fname2, tmp);
			fname1[0] = '\0';
			return;
		}
	}
	fname1[0] = '\0';
	fname2[0] = '\0';
	g_print("XXX: parse_files: wrong format\n %s", buf);
}
