/*
 * Module for file view(display "diff" result) in two-panes mode
 *
 * Programmed by INOUE Seiichiro <inoue@ainet.or.jp>
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#if defined(HAVE_ERRNO_H)
#include <errno.h>
#else defined(HAVE_SYS_ERRNO_H)
#include <sys/errno.h>
#endif
#include <gtk/gtk.h>
#include "gdiff.h"
#include "gui.h"
#include "misc.h"


/* Private function declarations */
static void init_state(GDiffFileViews *gfileviews);
static void create_textw(GDiffFileViews *gfileviews);
static void draw_text(GDiffFileViews *gfileviews, int whichfile);
static void show_hide_numbers(GDiffFileViews *gfileviews, int whichfile, gboolean b_show);
static void create_overview(GDiffFileViews *gfileviews);
static void draw_overview(GDiffFileViews *gfileviews, int whichfile);
static void draw_overview_lines(GtkWidget *widget, GdkEventExpose *event, gpointer data);
static void create_vscrollbar(GDiffFileViews *gfileviews);
static void vscrollbar_vchanged(GtkAdjustment *wadj, gpointer data);
/*
static void move_next(GDiffFileViews *gfileviews);
static void move_prev(GDiffFileViews *gfileviews);
*/


/**
 * create_twopane_widgets:
 * Called from gdiff_fileviews_new().
 * Create widgets and show them.
 * The contents(diff results) are not shown here, which is up to two_pane_display().
 * Input:
 * GDiffFileViews *gfileviews;
 * Output:
 * GDiffFileViews *gfileviews; updated.
 */
void
create_twopane_widgets(GDiffFileViews *gfileviews)
{
	GDiffWindow *gdwin = gfileviews->gdwin;
	GtkWidget *hbox;
	GtkWidget *label;
	gint page;
	int i;
	
	/* hbox */
	hbox = gtk_hbox_new(FALSE, 0);
	gfileviews->pane.two.hbox = GTK_HBOX(hbox);
	gfileviews->base = GTK_WIDGET(gfileviews->pane.two.hbox);/* Virtual base widget. */

	/* Add to the notebook */
	label = make_label_for_notebook(gfileviews->filename1, gfileviews->filename2);
	gfileviews->pane.two.label = GTK_LABEL(label);
	gtk_notebook_append_page(gdwin->notebook, hbox, label);

	/* text area */
	create_textw(gfileviews);
	/* overview area */
	create_overview(gfileviews);
	
	gtk_widget_show(hbox);
	/* Is this a proper way ? */
	page = gtk_notebook_page_num(gdwin->notebook, hbox);
	gtk_notebook_set_page(gdwin->notebook, page);

	create_rightclick_menu(gfileviews);
	
	/*XXX color. See "onepane.c" */
	for (i = 0; i < NUM_COMPARE_FILES; i++) {
		gfileviews->fcolor[i] = &(GTK_WIDGET(gfileviews->pane.two.overview[i])->style->fg[GTK_STATE_NORMAL]);
	}
}


/**
 * two_pane_display:
 * Show the diff result in two-panes mode.
 * Input:
 * GDiffFileViews *gfileviews;
 * Output:
 * GDiffFileViews *gfileviews; updated.
 */
void
two_pane_display(GDiffFileViews *gfileviews)
{
	DiffFiles *files = gfileviews->dfiles;
	FilesSpecialStatus fsst;
	int i;

	fsst = get_files_special_status(files);
	if (fsst != DIFFERENT_FILES) {
		g_message("binary or identical or only one file exists.(%d)\n", fsst);
		return;
	}
	init_state(gfileviews);
	for (i = 0; i < NUM_COMPARE_FILES; i++) {
		draw_text(gfileviews, i);
		draw_overview(gfileviews, i);
		if (gfileviews->state.show_line_num == TRUE) {
			show_hide_numbers(gfileviews, i, TRUE);
		}
	}
	/* to control both part */
	create_vscrollbar(gfileviews);
	
	/* Set right-click menu */
	for (i = 0; i < NUM_COMPARE_FILES; i++) {
		GtkText *text;
		text = gfileviews->pane.two.text[i];
		gtk_signal_connect(GTK_OBJECT(text), "button_press_event",
						   GTK_SIGNAL_FUNC(rightclick_menu_cb), gfileviews);
		gtk_object_set_user_data(GTK_OBJECT(text), (gpointer)i);
	}
}

/**
 * two_pane_show_linenum:
 * Show(Hide) the line numbers on text widget.
 * Input:
 * GDiffFileViews *gfileviews;
 * gboolean to_show; If TRUE, show. if FALSE, hide.
 * Output:
 * GDiffFileViews *gfileviews; updated.
 */
void
two_pane_show_linenum(GDiffFileViews *gfileviews, gboolean to_show)
{
	int i;

	if (to_show == TRUE && gfileviews->state.show_line_num == FALSE) {
		for (i = 0; i < NUM_COMPARE_FILES; i++) {
			show_hide_numbers(gfileviews, i, TRUE);
		}
		gfileviews->state.show_line_num = TRUE;
	} else if (to_show == FALSE && gfileviews->state.show_line_num == TRUE) {
		for (i = 0; i < NUM_COMPARE_FILES; i++) {
			show_hide_numbers(gfileviews, i, FALSE);
		}
		gfileviews->state.show_line_num = FALSE;
	}	
}

#if 0
/**
 * two_pane_next_prev:
 */ 
void
two_pane_next_prev(GDiffFileViews *gfileviews, gboolean to_next)
{
	if (to_next == TRUE)
		move_next(gfileviews);
	else
		move_prev(gfileviews);
}
#endif


/* ---The followings are private functions--- */
/**
 * init_state:
 */
static void init_state(GDiffFileViews *gfileviews)
{
	gfileviews->state.show_line_num = g_pref.show_line_num;
}

/**
 * create_textw:
 * Create GtkText widget.
 */
static void
create_textw(GDiffFileViews *gfileviews)
{
	GtkHBox *hbox = gfileviews->pane.two.hbox;
	int i;
	
	for (i = 0; i < NUM_COMPARE_FILES; i++) {
		GtkWidget *text;
		GtkWidget *scrolled_window;
		
		scrolled_window = gtk_scrolled_window_new(NULL, NULL);
		gtk_box_pack_start(GTK_BOX(hbox), scrolled_window, TRUE, TRUE, 0);
		gtk_widget_show(scrolled_window);
	
		text = gtk_text_new(NULL, NULL);
		gtk_text_set_editable(GTK_TEXT(text), FALSE);
		gtk_text_set_word_wrap(GTK_TEXT(text), FALSE);
		gtk_text_set_line_wrap(GTK_TEXT(text), g_pref.line_wrap);
		gtk_container_add(GTK_CONTAINER(scrolled_window), text);
		/*XXX gtk_widget_grab_focus(text);*/
		gtk_widget_show(text);

		gfileviews->pane.two.text[i] = GTK_TEXT(text);
	}
}

static void
draw_text(GDiffFileViews *gfileviews, int whichfile)
{
	DiffFiles *files = gfileviews->dfiles;
	const FileInfo *fi = get_fileinfo_from_diff_files(files, whichfile, TRUE);
	GtkText *text = gfileviews->pane.two.text[whichfile];
	GdkFont *font = GTK_WIDGET(text)->style->font;
	const char *pt;
	const char *text_last;/* for calculation of the length left */
	int line;
	GSList *node;/* node of DiffLines' list */
	FontProp fprop;
	GdkColor fcolor_common = {0, 0xFFFF, 0xFFFF, 0xFFFF};/* white */
	GdkColor *fcolor = gfileviews->fcolor[whichfile];

	fprop.font = font;
	fprop.fg = NULL;

	line = 1;
	pt = fi->text;
	text_last = fi->text + fi->lenb;
	gtk_text_freeze(text);
	for (node = files->dlines_list; node; node = node->next) {
		DiffLines *dlines;
		int begin, end;
		int lenb;
		
		dlines = (DiffLines*)(node->data);
		/* Use local variables for ease to read */
		begin = dlines->between[whichfile].begin;
		end = dlines->between[whichfile].end;

		if (end == 0)
			continue;
		fprop.bg = &fcolor_common;
		lenb = draw_text_lines(text, &fprop, pt, begin - line, text_last - pt);
		pt += lenb;

		if ((dlines->difftype == CHANGE)
			|| (dlines->difftype == F1ONLY && whichfile == FIRST_FILE)
			|| (dlines->difftype == F2ONLY && whichfile == SECOND_FILE)) {
			fprop.bg = fcolor;
		} else if ((dlines->difftype == F2ONLY && whichfile == FIRST_FILE)
				   || (dlines->difftype == F1ONLY && whichfile == SECOND_FILE)) {
			fprop.bg = &fcolor_common;
		}
		lenb = draw_text_lines(text, &fprop, pt, end - begin + 1, text_last - pt);
		pt += lenb;
		line = end + 1;
	}
	/* Draw the remained part */
	if (line >= fi->nline) {
		g_print("Problem?\n");
	}
	fprop.bg = &fcolor_common;
	draw_text_lines(text, &fprop, pt, fi->nline - line + 1, text_last - pt);
	gtk_text_thaw(text);
}


/* line numbers */
static void
show_hide_numbers(GDiffFileViews *gfileviews, int whichfile, gboolean b_show)
{
	DiffFiles *files = gfileviews->dfiles;
	const FileInfo *fi = get_fileinfo_from_diff_files(files, whichfile, TRUE);
	GtkText *text = gfileviews->pane.two.text[whichfile];
	GdkFont *font = GTK_WIDGET(text)->style->font;
	const char *pt;
	const char *text_last;/* for calculation of the length left */
	int line;
	int n_col;/* the number of columns for representing line numbers */
	const char *p_mark;
	char mark[MARK_LENGTH+1];
	int pos = 0;/* position in text widget */
	GSList *node;/* node of DiffLines' list */
	FontProp fprop;
	GdkColor fcolor_common = {0, 0xFFFF, 0xFFFF, 0xFFFF};/* white */
	GdkColor *fcolor = gfileviews->fcolor[whichfile];

	fprop.font = font;
	fprop.fg = NULL;

	line = 1;
	pt = fi->text;
	text_last = fi->text + fi->lenb;
	n_col = calc_number_places(fi->nline);
	if (whichfile == FIRST_FILE)
		strcpy(mark, MARK_FILE1);
	else
		strcpy(mark, MARK_FILE2);
	gtk_text_freeze(text);
	for (node = files->dlines_list; node; node = node->next) {
		DiffLines *dlines;
		int begin, end;
		int lenb;
		
		dlines = (DiffLines*)(node->data);
		/* Use local variables for ease to read */
		begin = dlines->between[whichfile].begin;
		end = dlines->between[whichfile].end;
		if (end == 0)
			continue;

		fprop.bg = &fcolor_common;
		p_mark = MARK_COMMON;
		lenb = insert_remove_line_numbers(text, b_show, &pos, &fprop, pt,
										  line, begin - line, text_last - pt,
										  n_col, p_mark);
		pt += lenb;
		if ((dlines->difftype == CHANGE)
			|| (dlines->difftype == F1ONLY && whichfile == FIRST_FILE)
			|| (dlines->difftype == F2ONLY && whichfile == SECOND_FILE)) {
			fprop.bg = fcolor;
			p_mark = mark;
		} else if ((dlines->difftype == F2ONLY && whichfile == FIRST_FILE)
				   || (dlines->difftype == F1ONLY && whichfile == SECOND_FILE)) {
			fprop.bg = &fcolor_common;
			p_mark = MARK_COMMON;
		}
		lenb = insert_remove_line_numbers(text, b_show, &pos, &fprop, pt,
										  begin, end - begin + 1, text_last - pt,
										  n_col, p_mark);
		pt += lenb;
		line = end + 1;
	}
	/* Draw the remained part */
	fprop.bg = &fcolor_common;
	p_mark = MARK_COMMON;
	insert_remove_line_numbers(text, b_show, &pos, &fprop, pt,
							   line, fi->nline - line + 1, text_last - pt,
							   n_col, p_mark);
	gtk_text_thaw(text);
}


/* overview */
/**
 * create_overview:
 * Create GdiffOverview widgets. See "gdiffoverview.[ch]".
 */
static void
create_overview(GDiffFileViews *gfileviews)
{
	GtkHBox *hbox = gfileviews->pane.two.hbox;
	GtkAdjustment *vadj;
	GtkWidget *ov1;
	GtkWidget *ov2;
	GtkWidget *darea;/* drawing area */

	
	darea = gtk_drawing_area_new();
	gtk_widget_set_name(GTK_WIDGET(darea), "drawing overview");
	vadj = GTK_TEXT(gfileviews->pane.two.text[FIRST_FILE])->vadj;
	ov1 = gdiff_overview_new(vadj);
	gtk_widget_set_name(GTK_WIDGET(ov1), "file1 overview");
	vadj = GTK_TEXT(gfileviews->pane.two.text[SECOND_FILE])->vadj;
	ov2 = gdiff_overview_new(vadj);
	gtk_widget_set_name(GTK_WIDGET(ov2), "file2 overview");

	gtk_box_pack_start(GTK_BOX(hbox), ov1, FALSE, FALSE, 0);
	gtk_widget_show(ov1);
	gtk_box_pack_start(GTK_BOX(hbox), darea, FALSE, FALSE, 0);
	gtk_widget_show(darea);
	gtk_box_pack_start(GTK_BOX(hbox), ov2, FALSE, FALSE, 0);
	gtk_widget_show(ov2);
	
	/* adjust drawing-area */
	gtk_drawing_area_size(GTK_DRAWING_AREA(darea), WIDTH_OVERVIEW, 0);/*XXX*/
	gdiff_overview_size(GDIFF_OVERVIEW(ov1), WIDTH_OVERVIEW, 0);/*XXX*/
	gdiff_overview_size(GDIFF_OVERVIEW(ov2), WIDTH_OVERVIEW, 0);/*XXX*/
	
	gfileviews->pane.two.darea = GTK_DRAWING_AREA(darea);
	gfileviews->pane.two.overview[FIRST_FILE] = GDIFF_OVERVIEW(ov1);
	gfileviews->pane.two.overview[SECOND_FILE] = GDIFF_OVERVIEW(ov2);
}


static void
draw_overview(GDiffFileViews *gfileviews, int whichfile)
{
	DiffFiles *files = gfileviews->dfiles;
	const FileInfo *fi = get_fileinfo_from_diff_files(files, whichfile, TRUE);
	GdiffOverview *ov = gfileviews->pane.two.overview[whichfile];
	GtkDrawingArea *darea = gfileviews->pane.two.darea;
	GSList *node;/* node of DiffLines' list */
	int all_nline;

	all_nline = fi->nline;
	if (all_nline == 0)
		return;

	for (node = files->dlines_list; node; node = node->next) {
		DiffLines *dlines;
		int begin, end;
		
		dlines = (DiffLines*)(node->data);
		/* Use local variables for ease to read */
		begin = dlines->between[whichfile].begin;
		end = dlines->between[whichfile].end;

		if ((dlines->difftype == CHANGE)
			|| (dlines->difftype == F1ONLY && whichfile == FIRST_FILE)
			|| (dlines->difftype == F2ONLY && whichfile == SECOND_FILE)) {
			gdiff_overview_insert_paintrange(ov, (gdouble)begin/all_nline, (gdouble)end/all_nline);
		} else if ((dlines->difftype == F2ONLY && whichfile == FIRST_FILE)
				   || (dlines->difftype == F1ONLY && whichfile == SECOND_FILE)) {
			/*XXX:*/
			gdiff_overview_insert_paintrange(ov, (gdouble)begin/all_nline, (gdouble)end/all_nline);
		}
	}
	/* Overview drawing area */
	gtk_signal_connect(GTK_OBJECT(darea), "expose_event",
                      (GtkSignalFunc)draw_overview_lines, gfileviews);
	
	return;
}

static void
draw_overview_lines(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
	GDiffFileViews *gfileviews = (GDiffFileViews*)data;
	DiffFiles *files = gfileviews->dfiles;
	const FileInfo *fi1 = get_fileinfo_from_diff_files(files, FIRST_FILE, TRUE);
	const FileInfo *fi2 = get_fileinfo_from_diff_files(files, SECOND_FILE, TRUE);
	int all_nline1 = fi1->nline;
	int all_nline2 = fi2->nline;
	GSList *node;/* node of DiffLines' list */
	GdkWindow *w = widget->window;
	GdkGC *gc = widget->style->fg_gc[GTK_STATE_NORMAL];
	int darea_width;
	int darea_height;

	gdk_window_get_size(w, &darea_width, &darea_height);
	for (node = files->dlines_list; node; node = node->next) {
		int y1, y2;
		DiffLines *dlines;
		int begin1, end1;
		int begin2, end2;

		dlines = (DiffLines*)(node->data);
		/* Use local variables for ease to read */
		begin1 = dlines->between[FIRST_FILE].begin;
		end1 = dlines->between[FIRST_FILE].end;
		begin2 = dlines->between[SECOND_FILE].begin;
		end2 = dlines->between[SECOND_FILE].end;

		if (all_nline1 != 0)
			y1 = darea_height * ((double)(begin1+end1)/(2*all_nline1));
		else
			y1 = 0;
		if (all_nline2 != 0)
			y2 = darea_height * ((double)(begin2+end2)/(2*all_nline2));
		else
			y2 = 0;
		gdk_draw_line(w, gc, 0, y1, darea_width, y2);
	}
}


/* Scrollbar to control both part */
static void
create_vscrollbar(GDiffFileViews *gfileviews)
{
	GtkHBox *hbox = gfileviews->pane.two.hbox;
	GtkWidget *vscrollbar;
	GtkAdjustment *adj;
	int max_nlines;

	max_nlines = get_max_nlines(gfileviews->dfiles);
	adj = GTK_ADJUSTMENT(gtk_adjustment_new(0, 0, max_nlines, 1, max_nlines/10.0, 1));
	vscrollbar = gtk_vscrollbar_new(adj);
	gtk_box_pack_start(GTK_BOX(hbox), vscrollbar, FALSE, FALSE, 0);
	gtk_signal_connect(GTK_OBJECT(adj), "value_changed",
					   GTK_SIGNAL_FUNC(vscrollbar_vchanged), gfileviews);
	gtk_widget_show(vscrollbar);
	gfileviews->pane.two.vscrollboth = GTK_VSCROLLBAR(vscrollbar);
}

static void
vscrollbar_vchanged(GtkAdjustment *wadj, gpointer data)
{
	GDiffFileViews *gfileviews = (GDiffFileViews*)data;
	GtkAdjustment *adj1 = gfileviews->pane.two.text[FIRST_FILE]->vadj;
	GtkAdjustment *adj2 = gfileviews->pane.two.text[SECOND_FILE]->vadj;
	gfloat max_nlines = wadj->upper;
	gfloat f1_max = adj1->upper - adj1->lower - adj1->page_size;
	gfloat f2_max = adj2->upper - adj2->lower - adj2->page_size;
	gfloat value = wadj->value;

	gtk_adjustment_set_value(adj1, value * f1_max / max_nlines);
	gtk_adjustment_set_value(adj2, value * f2_max / max_nlines);
}


#if 0
/* Move next or previous part */
static void
move_next(GDiffFileViews *gfileviews)
{
	DiffFiles *files = gfileviews->dfiles;
//	const FileInfo *fi = get_fileinfo_from_diff_files(files, FIRST_FILE, TRUE);
//	GtkText *text = gfileviews->pane.two.text[FIRST_FILE];
	int cur_line;

//	cur_line = get_current_line(vadj, fi->nline);
	g_print("cur %d\n", cur_line);
	
}

static void
move_prev(GDiffFileViews *gfileviews)
{
}
#endif
